import pathlib
from typing import List

from fhir.resources.fhirresourcemodel import FHIRResourceModel
from fhir.resources.coding import Coding
from fhir.resources.identifier import Identifier
from fhir.resources.reference import Reference
from dataclasses import dataclass
from urllib.parse import urlparse

from aced_submission.util import directory_reader


@dataclass
class ValidateDirectoryResult:
    """Results of FHIR validation of directory."""
    resources: List[FHIRResourceModel]
    exceptions: List[Exception]


def _check_coding(self: Coding, *args, **kwargs):
    """MonkeyPatch replacement for dict(), check Coding."""
    # note `self` is the Coding
    assert self.code, f"Missing `code` {self}"
    assert (not self.code.startswith("http")), f"`code` should _not_ be a url http {self.code}"
    assert ":" not in self.code, f"`code` should not contain ':' {self.code}"
    assert self.system, f"Missing `system` {self}"
    assert "%" not in self.system, f"`system` should be a simple url without uuencoding {self.system}"
    parsed = urlparse(self.system)
    assert parsed.scheme, f"`system` is not a URI {self}"
    assert self.display, f"Missing `display` {self}"
    # call the original dict() method
    return orig_coding_dict(self, *args, **kwargs)


def _check_identifier(self: Identifier, *args, **kwargs):
    """MonkeyPatch replacement for dict(), check Identifier."""
    # note `self` is the Identifier
    assert self.value, f"Missing `value` {self}"
    assert self.system, f"Missing `system` {self}"
    parsed = urlparse(self.system)
    assert parsed.scheme, f"`system` is not a URI {self}"
    assert "%" not in self.system, f"`system` should be a simple url without uuencoding {self.system}"
    # call the original dict() method
    return orig_identifier_dict(self, *args, **kwargs)


def _check_reference(self: Reference, *args, **kwargs):
    """MonkeyPatch replacement for dict(), check Reference."""
    # note `self` is the Identifier
    assert self.reference, f"Missing `reference` {self}"
    assert '/' in self.reference, f"Does not appear to be Relative reference {self}"
    assert 'http' not in self.reference, f"Absolute references not supported {self}"
    assert len(self.reference.split('/')) == 2, f"Does not appear to be Relative reference {self}"

    # call the original dict() method
    return orig_reference_dict(self, *args, **kwargs)


def validate_directory(directory_path: pathlib.Path) -> ValidateDirectoryResult:
    """Check FHIR data, accumulate results."""
    exceptions = []
    resources = []
    for parse_result in directory_reader(directory_path):
        if parse_result.exception:
            exceptions.append(parse_result)
        else:
            resources.append(parse_result.resource)
    return ValidateDirectoryResult(resources=resources, exceptions=exceptions)


#
# monkey patch dict() methods
#
orig_coding_dict = Coding.dict
Coding.dict = _check_coding

orig_identifier_dict = Identifier.dict
Identifier.dict = _check_identifier

orig_reference_dict = Reference.dict
Reference.dict = _check_reference



# https://github.com/uc-cdis/pelican/blob/master/pelican/dictionary.py#L7

import itertools
from collections import defaultdict


class DataDictionaryTraversal:
    def __init__(self, model):
        self.model = model

    def get_nodes(self):
        return self.model.Node.__subclasses__()

    def get_edges(self):
        return self.model.Edge.__subclasses__()

    def get_node_table_by_label(self):
        nodes = self.get_nodes()
        node_tables = {str(node.label): node.__tablename__ for node in nodes}
        return node_tables

    def get_node_label_by_table(self):
        nodes = self.get_nodes()
        node_tables = {node.__tablename__: str(node.label) for node in nodes}
        return node_tables

    def get_edge_table_by_labels(self):
        edges = self.get_edges()
        edge_tables = {
            (
                self.model.Node.get_subclass_named(edge.__src_class__).label,
                self.model.Node.get_subclass_named(edge.__dst_class__).label,
            ): edge.__tablename__
            for edge in edges
        }
        return edge_tables

    def get_edge_labels_by_table(self):
        edges = self.get_edges()
        edge_tables = {
            edge.__tablename__: {
                "src": self.model.Node.get_subclass_named(edge.__src_class__).label,
                "dst": self.model.Node.get_subclass_named(edge.__dst_class__).label,
            }
            for edge in edges
        }
        return edge_tables

    def get_edges_by_node(self):
        edges = self.get_edges()
        it = defaultdict(list)

        for edge in edges:
            it[self.model.Node.get_subclass_named(edge.__src_class__).label].append(
                edge.__tablename__
            )

        return it

    def _get_bfs(self, node_name):
        queue = [node_name]

        visited = {}

        r = []

        while queue:
            s = queue.pop(0)

            node = self.model.Node.get_subclass(s).__name__
            edges = self.model.Edge._get_edges_with_dst(node)

            r.append(s)

            for i in [
                self.model.Node.get_subclass_named(e.__src_class__).get_label()
                for e in edges
            ]:
                if i not in visited:
                    queue.append(i)
                    visited[i] = True

        return r

    def _topology_order(self, node_name, source_edges, target_class):
        stack, path = [node_name], []

        while stack:
            vertex = stack[-1]

            node_class = self.model.Node.get_subclass(vertex)
            assert (
                node_class
            ), f"Node name '{vertex}' does not exist in the graph data model. Maybe you need to provide a 'root_node' (see Pelican documentation)?"
            node = node_class.__name__
            edges = getattr(self.model.Edge, source_edges)(node)

            children = [
                self.model.Node.get_subclass_named(getattr(e, target_class)).get_label()
                for e in edges
            ]

            visited_children = [child for child in children if child not in path]

            if not visited_children:
                path.insert(0, vertex)
                stack.pop()
            else:
                stack.append(visited_children[0])

        return path

    def get_upward_path(self, node_name):
        return self._topology_order(node_name, "_get_edges_with_src", "__dst_class__")

    def get_downward_path(self, node_name):
        return self._topology_order(node_name, "_get_edges_with_dst", "__src_class__")

    def full_traverse_path(self, node_name, extra_nodes=None, include_upward=False):
        if include_upward:
            upward_path = list(
                zip(itertools.repeat(False), self.get_upward_path(node_name))
            )
            downward_path = list(
                zip(itertools.repeat(True), self.get_downward_path(node_name))
            )[1:]
            if extra_nodes:
                path = (
                    upward_path
                    + list(zip(itertools.repeat(True), extra_nodes))
                    + downward_path
                )
            else:
                path = upward_path + downward_path
        else:
            downward_path = list(
                zip(itertools.repeat(True), self.get_downward_path(node_name))
            )
            path = downward_path

        return path



"""Run flake8 tests"""

import subprocess
from typing import List


def test_coding_conventions(python_source_directories: List[str]):
    """Check python conventions on key directories"""
    failures = []
    for directory in python_source_directories:
        cmd_str = f"flake8 {directory}"
        completed = subprocess.run(cmd_str, shell=True)
        if completed.returncode != 0:
            _ = f"FAILURE: Python formatting and style for directory {directory}/"
            failures.append(_)
            print(_)

    assert len(failures) == 0, failures



import json
import pathlib
import sys
import threading
import uuid

import click
import yaml
from fhir.resources import FHIRAbstractModel  # noqa

from aced_submission import NaturalOrderGroup
from aced_submission.meta_graph_load import meta_upload, _table_mappings, _connect_to_postgres

from aced_submission.meta_flat_load import cli as meta_flat_load_cli
from aced_submission.meta_discovery_load import discovery_load

LINKS = threading.local()
CLASSES = threading.local()
IDENTIFIER_LIST_SIZE = 8

ACED_NAMESPACE = uuid.uuid3(uuid.NAMESPACE_DNS, 'aced-ipd.org')


@click.group(cls=NaturalOrderGroup)
def meta():
    """Project data (ResearchStudy, ResearchSubjects, Patient, etc.)."""
    pass


@meta.group(name='graph')
def graph():
    """Gen3 Graph data (nodes, edges, etc.)."""
    pass


@graph.command(name='upload')
@click.option('--source_path', required=False, default=None, show_default=True,
              help='Path on local file system')
@click.option('--project_id', required=True, show_default=True,
              help='Gen3 program-project')
@click.option('--credentials_file', default='~/.gen3/credentials.json', show_default=True,
              help='API credentials file downloaded from gen3 profile.')
@click.option('--silent', default=False, is_flag=True, show_default=True,
              help="No progress bar, or other output")
@click.option('--dictionary_path',
              default=None,  # 'output/gen3',
              show_default=True,
              help='Path to data dictionary (path or url).')
@click.option('--config_path',
              default='config.yaml',
              show_default=True,
              help='Path to config file.')
def _meta_upload(source_path, project_id, credentials_file, silent, dictionary_path, config_path):
    """Copy simplified json into sheepdog database."""
    program, project = project_id.split('-')
    assert program, "program is required"
    assert project, "project is required"

    meta_upload(source_path, program, project, credentials_file, silent, dictionary_path, config_path,
                file_name_pattern='**/*.ndjson')


@graph.command(name='mapping')
@click.option('--dictionary_path',
              default=None,  # 'output/gen3',
              show_default=True,
              help='Path to data dictionary (path or url).')
@click.option('--config_path',
              default='config.yaml',
              show_default=True,
              help='Path to config file.')
@click.option('--format', 'output_format',
              default='yaml',
              show_default=True,
              type=click.Choice(['yaml', 'json'], case_sensitive=False),
              )
def _meta_mapping(dictionary_path, config_path, output_format):
    """Show the table mappings for a data dictionary."""
    assert dictionary_path, '--dictionary_path is required'
    dictionary_dir = dictionary_path if 'http' not in dictionary_path else None
    dictionary_url = dictionary_path if 'http' in dictionary_path else None
    mappings = [mapping for mapping in _table_mappings(dictionary_dir, dictionary_url)]
    if output_format == 'yaml':
        yaml.dump(mappings, sys.stdout, default_flow_style=False)
    else:
        json.dump(mappings, sys.stdout, indent=2)


@graph.command(name='counts')
@click.option('--config_path',
              default='config.yaml',
              show_default=True,
              help='Path to config file.')
@click.option('--format', 'output_format',
              default='yaml',
              show_default=True,
              type=click.Choice(['yaml', 'json'], case_sensitive=False),
              )
def _node_counts(config_path, output_format):
    """Node counts in the sheepdog database by project."""
    assert config_path, '--config_path is required'
    assert pathlib.Path(config_path).exists(), f'--config_path {config_path} does not exist'
    dependency_order = yaml.safe_load(open(config_path))['dependency_order']
    dependency_order = [_ for _ in dependency_order if not _.startswith('_') and _ not in ['Project', 'Program']]
    dependency_order = [(rank + 1, _) for rank, _ in enumerate(dependency_order)]

    sql_selects = [
        "select 0 as hierarchy_rank, 'node_project' as table, 'Project' as node, _props->>'code' as project, count(*) as count from node_project group by _props->>'code'"]
    for rank, _ in dependency_order:
        if _.startswith('_'):
            continue
        sql_selects.append(
            f"select {rank} as hierarchy_rank, 'node_{_.lower()}' as table, '{_}' as node , _props->>'project_id' as project, count(*) as count from node_{_.lower()} group by _props->>'project_id'")

    sql = "\nunion\n".join(sql_selects) + "ORDER BY hierarchy_rank;"
    conn = _connect_to_postgres()

    # check program/project exist
    cur = conn.cursor()
    cur.execute(sql)
    counts = cur.fetchall()
    counts = [{'hierarchy_rank': _[0], 'table': _[1], 'node': _[2], 'project_id': _[3], 'count': _[4]} for _ in counts]
    if output_format == 'yaml':
        yaml.dump(counts, sys.stdout, default_flow_style=False)
    else:
        json.dump(counts, sys.stdout, indent=2)


@graph.command(name='rm')
@click.option('--config_path',
              default='config.yaml',
              show_default=True,
              help='Path to config file.')
@click.option('--project_id', required=True,
              default=None,
              show_default=True,
              help='program-project'
              )
@click.option('--format', 'output_format',
              default='yaml',
              show_default=True,
              type=click.Choice(['yaml', 'json'], case_sensitive=False),
              )
def _graph_rm(config_path, project_id, output_format):
    """Remove records from the sheepdog database by project."""
    assert config_path, '--config_path is required'
    assert pathlib.Path(config_path).exists(), f'--config_path {config_path} does not exist'

    program, project = project_id.split('-')
    assert program, "program is required"
    assert project, "project is required"

    reverse_dependency_order = yaml.safe_load(open(config_path))['dependency_order']
    reverse_dependency_order = [_ for _ in reverse_dependency_order if not _.startswith('_') and _ not in ['Project', 'Program']]
    reverse_dependency_order = [(rank + 1, _) for rank, _ in enumerate(reverse_dependency_order)]
    reverse_dependency_order.reverse()

    results = []
    conn = _connect_to_postgres()
    with conn:
        with conn.cursor() as curs:
            for rank, _ in reverse_dependency_order:
                if _.startswith('_'):
                    continue
                curs.execute(f"delete from node_{_.lower()} where _props->>'project_id' = %s;", (project_id,))
                results.append({'table': f'node_{_.lower()}', 'project_id': project_id, 'count': curs.rowcount})
            curs.execute("delete from node_project where _props->>'code' = %s;", (project,))
            results.append({'table': 'node_project', 'project_id': project, 'count': curs.rowcount})

    if output_format == 'yaml':
        yaml.dump(results, sys.stdout, default_flow_style=False)
    else:
        json.dump(results, sys.stdout, indent=2)


meta.add_command(meta_flat_load_cli)


@meta.group(name='discovery')
def discovery():
    """Gen3 discovery database."""
    pass


@discovery.command('load')
@click.option('--program', default="aced", show_default=True,
              help='Gen3 "program"')
@click.option('--credentials_file', default='~/.gen3/credentials.json', show_default=True,
              help='API credentials file downloaded from gen3 profile.')
def discovery(program, credentials_file):
    """Writes project information to discovery metadata-service"""
    discovery_load(program, credentials_file)


if __name__ == '__main__':
    meta()



import gzip
import importlib
import io
import pathlib
from dataclasses import dataclass
from typing import Dict, Iterator

import orjson
from fhir.resources.fhirresourcemodel import FHIRResourceModel
from pydantic import ValidationError
import logging

FHIR_CLASSES = importlib.import_module('fhir.resources')

logger = logging.getLogger(__name__)


@dataclass
class ParseResult:
    """Results of FHIR validation of dict."""
    resource: FHIRResourceModel
    """If valid, the FHIR resource."""
    exception: Exception
    """If invalid, the exception."""
    path: pathlib.Path = None
    """Source file, if available."""
    offset: int = None
    """Base 0 offset of line number(ndjson) or entry(bundle)."""
    resource_id: str = None
    """Resource id of resource"""


def parse_obj(resource: Dict, validate=True) -> ParseResult:
    """Load a dictionary into a FHIR model """
    try:
        assert 'resourceType' in resource, "Dict missing `resourceType`, is it a FHIR dict?"
        klass = FHIR_CLASSES.get_fhir_model_class(resource['resourceType'])
        _ = klass.parse_obj(resource)
        if validate:
            # trigger object traversal, see monkey patch below, at bottom of file
            _.dict()
        return ParseResult(resource=_, exception=None, path=None, resource_id=_.id)
    except (ValidationError, AssertionError) as e:
        return ParseResult(resource=None, exception=e, path=None, resource_id=resource.get('id', None))


def _is_ndjson(file_path: pathlib.Path) -> bool:
    """Open file, read all lines as json."""
    fp = _to_file(file_path)
    try:
        with fp:
            for line in fp.readlines():
                orjson.loads(line)
                break
        return True
    except Exception as e:  # noqa
        return False


def _to_file(file_path):
    """Open file appropriately."""
    if file_path.name.endswith('gz'):
        fp = io.TextIOWrapper(io.BufferedReader(gzip.GzipFile(file_path)))  # noqa
    else:
        fp = open(file_path, "rb")
    return fp


def _is_json_file(name: str) -> bool:
    """Files we are interested in"""
    if name.endswith('json.gz'):
        return True
    if name.endswith('json'):
        return True
    return False


def _has_entries(_: ParseResult):
    """"""
    if _.resource is None:
        return False
    return _.resource.resource_type in ["Bundle", "List"] and _.resource.entry is not None


def _entry_iterator(parse_result: ParseResult) -> Iterator[ParseResult]:
    """See if there are entries"""
    if not _has_entries(parse_result):
        yield parse_result
    else:
        _path = parse_result.path
        offset = 0
        if parse_result.resource.entry and len(parse_result.resource.entry) > 0:
            for _ in parse_result.resource.entry:
                if _ is None:
                    break
                if hasattr(_, 'resource'):  # BundleEntry
                    yield ParseResult(path=_path, resource=_.resource, offset=offset, exception=None)
                elif hasattr(_, 'item'):  # ListEntry
                    yield ParseResult(path=_path, resource=_.item, offset=offset, exception=None)
                else:
                    yield ParseResult(path=_path, resource=_.item, offset=offset, exception=None)
                offset += 1
    pass


def directory_reader(directory_path: pathlib.Path, pattern: str = '*.*', validate=True) -> Iterator[ParseResult]:
    """Extract FHIR resources from directory"""

    assert directory_path.is_dir(), f"{directory_path.name} is not a directory"

    input_files = [_ for _ in directory_path.glob(pattern) if _is_json_file(_.name)]
    for input_file in input_files:
        logger.info(input_file)
        if not input_file.is_file():
            continue
        is_ndjson = _is_ndjson(input_file)
        fp = _to_file(input_file)
        with fp:
            if is_ndjson:
                offset = 0
                for line in fp.readlines():
                    parse_result = parse_obj(orjson.loads(line), validate)
                    parse_result.path = input_file
                    parse_result.offset = offset
                    for _ in _entry_iterator(parse_result):
                        # print(_.offset, 'is_ndjson', input_file, parse_result.resource.id, parse_result.resource.resource_type)
                        yield _
                    offset += 1
            else:
                # look for json bundles
                _ = orjson.loads(fp.read())
                # not a bundle
                parse_result = parse_obj(_, validate)
                parse_result.path = input_file
                parse_result.offset = 0
                for _ in _entry_iterator(parse_result):
                    yield _
                continue



#  with a running Gen3
cmd = """
GEN3SDK_MAX_RETRIES=1 aced_submission files upload --document_reference_path tmp/aced-foo/DocumentReference.ndjson  --program aced --project Alcoholism --bucket_name aced-ohsu --duplicate_check  --ignore_state
"""  # TODO



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
"""

from os import path

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get the requirements from the requirements.txt file
with open(path.join(here, "requirements.txt"), encoding='utf-8') as f:
    requirements = [l_ for l_ in f.read().splitlines() if
                    not l_.startswith('#') and len(l_) > 0 and not l_.startswith('http')]

with open(path.join(here, "requirements.txt"), encoding='utf-8') as f:
    dependency_links = [l_ for l_ in f.read().splitlines() if l_.startswith('http')]

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install hop_etl
    #
    # And where it will live on PyPI: https://pypi.org/project/drs_downloader/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='aced_submission',  # Required

    # Versions should comply with PEP 440:
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.0.9rc2',  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='Populate ACED commons from FHIR resources',
    # Optional

    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional

    # Denotes that our long_description is in Markdown
    long_description_content_type='text/markdown',  # Optional (see note above)

    # Project's main homepage.
    url='https://github.com/ACED-IDP/aced_submission',  # Optional

    # This should be your name or the name of the organization which owns the
    # project.
    author='compbio',  # Optional


    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here.
        'Programming Language :: Python :: 3.9',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='gen3 bioinformatics',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required

    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. If you
    # do not support Python 2, you can simplify this to '>=3.5' or similar, see
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='>=3.6, <4',

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=requirements,

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    package_data={  # Optional
        '': ['*.yaml'],
    },

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.

    entry_points={
        'console_scripts': [
            'aced_submission = aced_submission.cli:cli'
        ],
    },

    include_package_data=True,

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/ACED-IDP/aced_submission/issues',
        'Source': 'https://github.com/ACED-IDP/aced_submission',
    },
)



"""Load flat indexes into elasticsearch."""


import csv
import json
import logging
import os
import pathlib
import sqlite3
import uuid
from datetime import datetime
from functools import lru_cache
from itertools import islice
from typing import Dict, Iterator

import click
import elasticsearch
import orjson
import requests
from dictionaryutils import DataDictionary
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
logging.getLogger('elasticsearch').setLevel(logging.WARNING)

DEFAULT_ELASTIC = "http://localhost:9200"

k8s_elastic = os.environ.get('GEN3_ELASTICSEARCH_MASTER_PORT', None)
if k8s_elastic:
    DEFAULT_ELASTIC = f"http://{k8s_elastic.replace('tcp://', '')}"

# TODO - fix me should be gen3.aced-idp.org but we need to coordinate with gitops.json
ES_INDEX_PREFIX = "gen3.aced.io"

ACED_NAMESPACE = uuid.uuid3(uuid.NAMESPACE_DNS, 'aced-ipd.org')


def create_id(key: str) -> str:
    """Create an idempotent ID from the input string."""
    return str(uuid.uuid5(ACED_NAMESPACE, key))


def read_ndjson(path: str) -> Iterator[Dict]:
    """Read ndjson file, load json line by line."""
    with open(path) as jsonfile:
        for l_ in jsonfile.readlines():
            yield json.loads(l_)


def read_tsv(path: str) -> Iterator[Dict]:
    """Read tsv file line by line."""
    with open(path) as tsv_file:
        reader = csv.DictReader(tsv_file, delimiter="\t")
        for row in reader:
            yield row


def create_index_from_source(_schema, _index, _type):
    """Given an ES source dict, create ES index."""
    mappings = {}
    if _type == 'file':
        # TODO fix me - we should have a index called document_reference not file
        properties = _schema['document_reference.yaml']['properties']
    else:
        properties = _schema[_type + '.yaml']['properties']
    for k, v in properties.items():
        if not isinstance(v, dict):
            continue
        if '$ref' in v:
            (ref_, ref_prop) = v['$ref'].split('#/')
            prop_type = _schema[ref_][ref_prop]['type']
        elif 'enum' in v:
            prop_type = 'string'
        elif 'oneOf' in v:
            prop_type = 'string'
        else:
            if 'type' not in v:
                print("?")
            if isinstance(v['type'], list):
                prop_type = v['type'][0]
            else:
                prop_type = v['type']

        if prop_type in ['string']:
            mappings[k] = {
                "type": "keyword"
            }
        elif prop_type in ['boolean']:
            mappings[k] = {
                "type": "keyword"
            }
        elif 'date' in prop_type:
            mappings[k] = {
                "type": "date"
            }
        elif 'array' in prop_type:
            mappings[k] = {
                "type": "keyword"
            }
        else:
            # naive, there are probably other types
            mappings[k] = {"type": "float"}
        # we have a patient centric index approach, all links include a `patient`
        mappings['patient_id'] = {"type": "keyword"}
        # patient fields copied to observation
        if _type == 'observation':
            mappings['us_core_race'] = {"type": "keyword"}
            mappings['address'] = {"type": "keyword"}
            mappings['gender'] = {"type": "keyword"}
            mappings['birthDate'] = {"type": "keyword"}
            mappings['us_core_ethnicity'] = {"type": "keyword"}
            mappings['address_orh_zip_designation_code'] = {"type": "keyword"}
            mappings['condition'] = {"type": "keyword"}
            mappings['condition_code'] = {"type": "keyword"}
            mappings['family_history_condition'] = {"type": "keyword"}
            mappings['family_history_condition_code'] = {"type": "keyword"}

    mappings['auth_resource_path'] = {"type": "keyword"}

    dynamic_templates = [
        {
            "strings": {
                "match_mapping_type": "string",
                "mapping": {
                    "type": "keyword"
                }
            }
        }
    ]

    return {
        # https://www.elastic.co/guide/en/elasticsearch/reference/current/dynamic.html#dynamic-parameters
        "mappings": {"properties": mappings, 'dynamic_templates': dynamic_templates}
    }


def write_array_aliases(doc_type, alias, elastic=DEFAULT_ELASTIC, name_space=ES_INDEX_PREFIX):
    """Write the array aliases."""
    # EXPECTED_ALIASES = {
    #     ".kibana_1": {
    #         "aliases": {
    #             ".kibana": {}
    #         }
    #     },
    #     "etl-array-config_0": {
    #         "aliases": {
    #             "etl-array-config": {},
    #             "etl_array-config": {},
    #             "time_2022-08-25T01:44:47.115494": {}
    #         }
    #     },
    #     "etl_0": {
    #         "aliases": {
    #             "etl": {},
    #             "time_2022-08-25T01:44:47.115494": {}
    #         }
    #     },
    #     "file-array-config_0": {
    #         "aliases": {
    #             "file-array-config": {},
    #             "file_array-config": {},
    #             "time_2022-08-25T01:44:47.115494": {}
    #         }
    #     },
    #     "file_0": {
    #         "aliases": {
    #             "file": {},
    #             "time_2022-08-25T01:44:47.115494": {}
    #         }
    #     }
    # }
    return {
        "method": 'POST',
        "url": f'{elastic}/_aliases',
        "json": {
            "actions": [
                {"add": {"index": f"{name_space}_{doc_type}-array-config_0",
                         "alias": f"{name_space}_array-config"}},
                {"add": {"index": f"{name_space}_{doc_type}-array-config_0",
                         "alias": f"{alias}_array-config"}}
            ]}
    }


def write_array_config(doc_type, alias, field_array, elastic=DEFAULT_ELASTIC, name_space=ES_INDEX_PREFIX):
    """Write the array config."""
    return {
        "method": 'PUT',
        "url": f'/{name_space}_{doc_type}-array-config_0/_doc/{alias}',
        "json": {"timestamp": datetime.now().isoformat(), "array": field_array}
    }


def write_alias_config(doc_type, alias, elastic=DEFAULT_ELASTIC, name_space=ES_INDEX_PREFIX):
    """Write the alias config."""
    return {
        "method": 'POST',
        "url": f'{elastic}/_aliases',
        "json": {"actions": [{"add": {"index": f"{name_space}_{doc_type}_0", "alias": alias}}]}
    }


def create_indexes(_schema, _index, doc_type, elastic=DEFAULT_ELASTIC):
    """Create the es indexes."""
    return {
        "method": 'PUT',
        "url": f'{elastic}/{_index}',
        "json": create_index_from_source(_schema, _index, doc_type),
        "index": _index,
    }


def write_sqlite(index, generator):
    """Write to sqlite"""
    connection = sqlite3.connect(f'{index}.sqlite')
    with connection:
        connection.execute(f'DROP table IF EXISTS {index}')
        connection.execute(f'CREATE TABLE if not exists {index} (id PRIMARY KEY, entity Text)')
        with connection:
            connection.executemany(f'insert into {index} values (?, ?)',
                                   [(entity['id'], orjson.dumps(entity).decode(),) for entity in generator])


def write_bulk_http(elastic, index, limit, doc_type, generator, schema):
    """Use efficient method to write to elastic"""
    counter = 0

    def _bulker(generator_, counter_=counter):
        for dict_ in generator_:
            if limit and counter_ > limit:
                break  # for testing
            yield {
                '_index': index,
                '_op_type': 'index',
                '_source': dict_,
                # use the id from the FHIR object to upsert information
                '_id': dict_['id']
            }
            counter_ += 1
            if counter_ % 10000 == 0:
                logger.info(f"{counter_} records written")
        logger.info(f"{counter_} records written")

    if schema:
        logger.info(f'Creating {doc_type} indices.')
        index_dict = create_indexes(schema, _index=index, doc_type=doc_type)

        try:
            elastic.indices.create(index=index_dict['index'], body=index_dict['json'])
        except Exception as e:
            if 'resource_already_exists_exception' in str(e):
                logger.debug(f"Index already exists. {index} {str(e)}")
                logger.debug("Continuing to load.")
            else:
                raise e

    logger.info(f'Writing bulk to {index} limit {limit}.')
    _ = bulk(client=elastic,
             actions=(d for d in _bulker(generator)),
             request_timeout=120)


def observation_generator(project_id, path) -> Iterator[Dict]:
    """Render guppy index for observation."""
    program, project = project_id.split('-')

    connection = sqlite3.connect('denormalized_patient.sqlite')

    for observation in read_ndjson(path):
        o_ = observation['object']

        o_['project_id'] = project_id
        o_["auth_resource_path"] = f"/programs/{program}/projects/{project}"
        for relation in observation['relations']:
            dst_name = relation['dst_name'].lower()
            dst_id = relation['dst_id']
            o_[f'{dst_name}_id'] = dst_id

        #
        for required_field in []:
            if required_field not in o_:
                o_[required_field] = None

        assert 'patient_id' in o_, observation

        denormalized_patient = fetch_denormalized_patient(connection, o_['patient_id'])
        condition, condition_coding, fh_condition, fh_condition_coding, patient = (
            denormalized_patient['condition'],
            denormalized_patient['condition_coding'],
            denormalized_patient['fh_condition'],
            denormalized_patient['fh_condition_coding'],
            denormalized_patient['patient'],
        )

        if patient:
            o_['us_core_race'] = patient.get('us_core_race', None)
            o_['address'] = patient.get('address', None)
            o_['gender'] = patient.get('gender', None)
            o_['birthDate'] = patient.get('birthDate', None)
            o_['us_core_ethnicity'] = patient.get('us_core_ethnicity', None)
            o_['address_orh_zip_designation_code'] = patient.get('address_orh_zip_designation_code', None)

            o_['condition'] = condition
            o_['condition_code'] = condition_coding
            o_['family_history_condition'] = fh_condition
            o_['family_history_condition_code'] = fh_condition_coding

        yield o_


@lru_cache(maxsize=1024 * 10)
def fetch_denormalized_patient(connection, patient_id):
    """Retrieve unique conditions and family history"""

    fh_condition = []
    fh_condition_coding = []
    condition = []
    condition_coding = []
    patient = None

    for row in connection.execute('select entity from patient where id = ? limit 1', (patient_id,)):
        patient = orjson.loads(row[0])
        break

    for row in connection.execute('select entity from family_history where patient_id = ? ', (patient_id,)):
        family_history = orjson.loads(row[0])
        for _ in family_history['condition']:
            if _ not in fh_condition:
                fh_condition.append(_)
        for _ in family_history['condition_coding']:
            if _ not in fh_condition_coding:
                fh_condition_coding.append(_)

    for row in connection.execute('select entity from condition where patient_id = ? ', (patient_id,)):
        condition_ = orjson.loads(row[0])
        if condition_['code'] not in condition:
            condition.append(condition_['code'])
            condition_coding.append(condition_['code_coding'])

    return {
        'patient': patient, 'condition': condition, 'condition_coding': condition_coding,
        'fh_condition': fh_condition, 'fh_condition_coding': fh_condition_coding
    }


def patient_generator(project_id, path) -> Iterator[Dict]:
    """Render guppy index for patient."""
    program, project = project_id.split('-')
    for patient in read_ndjson(path):
        p_ = patient['object']
        p_['id'] = patient['id']

        p_['project_id'] = project_id
        p_["auth_resource_path"] = f"/programs/{program}/projects/{project}"

        #
        for required_field in []:
            if required_field not in p_:
                p_[required_field] = None
        yield p_


def file_generator(project_id, path) -> Iterator[Dict]:
    """Render guppy index for file."""
    program, project = project_id.split('-')
    for file in read_ndjson(path):
        f_ = file['object']
        f_['id'] = file['id']

        f_['project_id'] = project_id
        f_["auth_resource_path"] = f"/programs/{program}/projects/{project}"

        for relation in file['relations']:
            dst_name = relation['dst_name'].lower()
            dst_id = relation['dst_id']
            f_[f'{dst_name}_id'] = dst_id

        #
        for required_field in []:
            if required_field not in f_:
                f_[required_field] = None
        yield f_


def setup_aliases(alias, doc_type, elastic, field_array, index):
    """Create the alias to the data index"""
    elastic.indices.put_alias(index, alias)
    # create a configuration index that guppy will read that describes the array fields
    # TODO - find a doc or code reference in guppy that explains how this is used
    alias_index = f'{ES_INDEX_PREFIX}_{doc_type}-array-config_0'
    try:
        mapping = {
            "mappings": {
                "properties": {
                    "timestamp": {"type": "date"},
                    "array": {"type": "keyword"},
                }
            }
        }
        elastic.indices.create(index=alias_index, body=mapping)
    except Exception as e:
        logger.warning(f"Could not create index. {index} {str(e)}")
        logger.warning("Continuing to load.")

    try:
        elastic.create(alias_index, id=alias,
                       body={"timestamp": datetime.now().isoformat(), "array": field_array})
    except elasticsearch.exceptions.ConflictError:
        pass
    elastic.indices.update_aliases(
        {"actions": [{"add": {"index": f"{ES_INDEX_PREFIX}_{doc_type}_0", "alias": alias}}]}
    )
    elastic.indices.update_aliases({
        "actions": [
            {"add": {"index": f"{ES_INDEX_PREFIX}_{doc_type}-array-config_0",
                     "alias": f"{ES_INDEX_PREFIX}_array-config"}},
            {"add": {"index": f"{ES_INDEX_PREFIX}_{doc_type}-array-config_0",
                     "alias": f"{doc_type}_array-config"}}
        ]}
    )


@click.group('flat')
def cli():
    """Load flat indexes into elasticsearch."""
    pass


def write_flat_file(output_path, index, doc_type, limit, generator, schema):
    """Write the flat model to a file."""
    counter_ = 0

    pathlib.Path(output_path).mkdir(parents=True, exist_ok=True)

    with open(f"{output_path}/{doc_type}.ndjson", "wb") as fp:
        for dict_ in generator:
            fp.write(
                orjson.dumps(
                    {
                        'id': dict_['id'],
                        'object': dict_,
                        'name': doc_type,
                        'relations': []
                    }
                )
            )
            fp.write(b'\n')

            counter_ += 1
            if counter_ % 10000 == 0:
                logger.info(f"{counter_} records written")
        logger.info(f"{counter_} records written")


@cli.command('denormalize-patient')
@click.option('--input_path', required=True,
              default=None,
              show_default=True,
              help='Path to flattened json'
              )
def _denormalize_patient(input_path):
    denormalize_patient(input_path)


def denormalize_patient(input_path):
    """Gather Patient, FamilyHistory, Condition into sqlite db."""

    path = pathlib.Path(input_path)

    def _load_vertex(file_name):
        """Get the object and patient id"""
        if not (path / file_name).is_file():
            return
        for _ in read_ndjson(path / file_name):
            patient_id = None
            if len(_['relations']) == 1 and _['relations'][0]['dst_name'] == 'Patient':
                patient_id = _['relations'][0]['dst_id']
            _ = _['object']
            _['id'] = _['id']
            if patient_id:
                _['patient_id'] = patient_id
            yield _

    connection = sqlite3.connect('denormalized_patient.sqlite')
    with connection:
        connection.execute('DROP table IF EXISTS patient')
        connection.execute('DROP table IF EXISTS family_history')
        connection.execute('DROP table IF EXISTS condition')
        connection.execute('CREATE TABLE if not exists patient (id PRIMARY KEY, entity Text)')
        connection.execute('CREATE TABLE if not exists family_history (id PRIMARY KEY, patient_id Text, entity Text)')
        connection.execute('CREATE TABLE if not exists condition (id PRIMARY KEY, patient_id Text, entity Text)')
    with connection:
        connection.executemany('insert into patient values (?, ?)',
                               [(entity['id'], orjson.dumps(entity).decode(),) for entity in
                                _load_vertex('Patient.ndjson')])
    with connection:
        connection.executemany('insert into family_history values (?, ?, ?)',
                               [(entity['id'], entity['patient_id'], orjson.dumps(entity).decode(),) for entity in
                                _load_vertex('FamilyMemberHistory.ndjson')])
    with connection:
        connection.executemany('insert into condition values (?, ?, ?)',
                               [(entity['id'], entity['patient_id'], orjson.dumps(entity).decode(),) for entity in
                                _load_vertex('Condition.ndjson')])
    with connection:
        connection.execute('CREATE INDEX if not exists condition_patient_id on condition(patient_id)')
        connection.execute('CREATE INDEX if not exists family_history_patient_id on condition(patient_id)')


@cli.command('load')
@click.option('--project_id', required=True,
              default=None,
              show_default=True,
              help='program-project'
              )
@click.option('--index', required=True,
              default=None,
              show_default=True,
              help='Elastic index name'
              )
@click.option('--path', required=True,
              default=None,
              show_default=True,
              help='Path to flattened json'
              )
@click.option('--elastic_url', default=DEFAULT_ELASTIC, show_default=True)
@click.option('--limit',
              default=None,
              show_default=True,
              help='Max number of rows per index.')
@click.option('--schema_path', required=True,
              default='generated-json-schema/aced.json',
              show_default=True,
              help='Path to gen3 schema json'
              )
@click.option('--output_path', required=False,
              default=None,
              show_default=True,
              help='Do not load elastic, write flat model to file instead'
              )
def _load_flat(project_id, index, path, limit, elastic_url, schema_path, output_path):
    """Gen3 Elastic Search data into guppy (patient, observation, files, etc.)."""
    load_flat(project_id, index, path, limit, elastic_url, schema_path, output_path)


def load_flat(project_id, index, path, limit, elastic_url, schema_path, output_path):
    # replaces tube_lite

    if limit:
        limit = int(limit)

    elastic = Elasticsearch([elastic_url], request_timeout=120)

    index = index.lower()

    if 'http' in schema_path:
        schema = requests.get(schema_path).json()
    else:
        schema = DataDictionary(local_file=schema_path).schema

    if index == 'patient':
        doc_type = 'patient'
        index = f"{ES_INDEX_PREFIX}_{doc_type}_0"
        alias = 'patient'
        field_array = [k for k, v in schema['patient.yaml']['properties'].items() if 'array' in v.get('type', {})]

        if not output_path:
            # create the index and write data into it.
            write_bulk_http(elastic=elastic, index=index, doc_type=doc_type, limit=limit,
                            generator=patient_generator(project_id, path), schema=schema)

            setup_aliases(alias, doc_type, elastic, field_array, index)
        else:
            # write file path
            write_flat_file(output_path=output_path, index=index, doc_type=doc_type, limit=limit,
                            generator=patient_generator(project_id, path), schema=schema)

    if index == 'observation':
        doc_type = 'observation'
        index = f"{ES_INDEX_PREFIX}_{doc_type}_0"
        alias = 'observation'
        field_array = [k for k, v in schema['observation.yaml']['properties'].items() if 'array' in v.get('type', {})]
        # field_array = ['data_format', 'data_type', '_file_id', 'medications', 'conditions']

        if not output_path:
            # create the index and write data into it.
            write_bulk_http(elastic=elastic, index=index, doc_type=doc_type, limit=limit,
                            generator=observation_generator(project_id, path), schema=schema)

            setup_aliases(alias, doc_type, elastic, field_array, index)
        else:
            # write file path
            write_flat_file(output_path=output_path, index=index, doc_type=doc_type, limit=limit,
                            generator=observation_generator(project_id, path), schema=schema)

    if index == 'file':
        doc_type = 'file'
        alias = 'file'
        index = f"{ES_INDEX_PREFIX}_{doc_type}_0"
        field_array = [k for k, v in schema['document_reference.yaml']['properties'].items() if
                       isinstance(v, dict) and 'array' in v.get('type',
                                                                {})]
        if not output_path:
            # create the index and write data into it.
            write_bulk_http(elastic=elastic, index=index, doc_type=doc_type, limit=limit,
                            generator=file_generator(project_id, path), schema=schema)

            setup_aliases(alias, doc_type, elastic, field_array, index)
        else:
            # write file path
            write_flat_file(output_path=output_path, index=index, doc_type=doc_type, limit=limit,
                            generator=file_generator(project_id, path), schema=schema)


def chunk(arr_range, arr_size):
    """Iterate in chunks."""
    arr_range = iter(arr_range)
    return iter(lambda: tuple(islice(arr_range, arr_size)), ())


@cli.command('counts')
@click.option('--project_id', required=True,
              default=None,
              show_default=True,
              help='program-project'
              )
def _counts(project_id):
    counts(project_id)


def counts(project_id):
    """Count the number of patients, observations, and files."""
    elastic = Elasticsearch([DEFAULT_ELASTIC], request_timeout=120)
    program, project = project_id.split('-')
    assert program, "program is required"
    assert project, "project is required"
    query = {
        "query": {
            "match": {
                "auth_resource_path": f"/programs/{program}/projects/{project}"
            }
        }
    }
    for index in ['patient', 'observation', 'file']:
        # index = f"{ES_INDEX_PREFIX}_{index}_0"
        print(index, elastic.count(index=index, body=query)['count'])


@cli.command('rm')
@click.option('--project_id', required=True,
              default=None,
              show_default=True,
              help='program-project'
              )
@click.option('--index', required=True,
              default=None,
              show_default=True,
              help='one of patient, observation, file'
              )
def _delete(project_id, index):
    delete(project_id, index)


def delete(project_id, index):
    """Delete items from elastic index for project_id."""
    elastic = Elasticsearch([DEFAULT_ELASTIC], request_timeout=120)
    assert project_id, "project_id is required"
    program, project = project_id.split('-')
    assert program, "program is required"
    assert project, "project is required"
    assert index, "index is required"
    query = {
        "query": {
            "match": {
                "auth_resource_path": f"/programs/{program}/projects/{project}"
            }
        }
    }
    print("deleting, waiting up to 5 min. for response")
    print(index, elastic.delete_by_query(index=index, body=query, timeout='5m'))


if __name__ == '__main__':
    cli()



import os
import pathlib
import logging
import requests
from gen3.auth import Gen3Auth
from gen3.metadata import Gen3Metadata


# copied from gen3_util/config/__init__.py
def ensure_auth(refresh_file: [pathlib.Path, str] = None, validate: bool = False) -> Gen3Auth:
    """Confirm connection to Gen3 using their conventions.

    Args:
        refresh_file (pathlib.Path): The file containing the downloaded JSON web token.
        validate: check the connection by getting a new token

    """

    try:
        if 'ACCESS_TOKEN' in os.environ:
            access_token = os.environ
            auth = Gen3Auth(refresh_file=f"accesstoken:///{access_token}")

        elif refresh_file:
            if isinstance(refresh_file, str):
                refresh_file = pathlib.Path(refresh_file)
            auth = Gen3Auth(refresh_file=refresh_file.name)

        else:
            auth = Gen3Auth()

        if validate:
            api_key = auth.refresh_access_token()
            assert api_key, "refresh_access_token failed"

    except (requests.exceptions.ConnectionError, AssertionError) as e:
        msg = ("Could not get access. "
               "See https://uc-cdis.github.io/gen3-user-doc/appendices/api-gen3/#credentials-to-query-the-api. "
               "Store the file in ~/.gen3/credentials.json or specify location with env GEN3_API_KEY "
               f"{e}")

        logging.getLogger(__name__).error(msg)
        raise AssertionError(msg)

    return auth


def discovery_load(program, gen3_credentials_file):
    """Writes project information to discovery metadata-service"""
    auth = ensure_auth(gen3_credentials_file)
    discovery_client = Gen3Metadata(auth.endpoint, auth)
    # TODO - read from some other, more dynamic source
    discovery_descriptions = """
Alcoholism~9300~Patients from 'Coherent Data Set' https://www.mdpi.com/2079-9292/11/8/1199/htm that were diagnosed with condition(s) of: Alcoholism.  Data hosted by: aced-ohsu~aced-ohsu
Alzheimers~45306~Patients from 'Coherent Data Set' https://www.mdpi.com/2079-9292/11/8/1199/htm that were diagnosed with condition(s) of: Alzheimer's, Familial Alzheimer's.  Data hosted by: aced-ucl~aced-ucl
Breast_Cancer~7105~Patients from 'Coherent Data Set' https://www.mdpi.com/2079-9292/11/8/1199/htm that were diagnosed with condition(s) of: Malignant neoplasm of breast (disorder).  Data hosted by: aced-manchester~aced-manchester
Colon_Cancer~25355~Patients from 'Coherent Data Set' https://www.mdpi.com/2079-9292/11/8/1199/htm that were diagnosed with condition(s) of: Malignant tumor of colon,  Polyp of colon.  Data hosted by: aced-stanford~aced-stanford
Diabetes~65051~Patients from 'Coherent Data Set' https://www.mdpi.com/2079-9292/11/8/1199/htm that were diagnosed with condition(s) of: Diabetes.  Data hosted by: aced-ucl~aced-ucl
Lung_Cancer~25355~Patients from 'Coherent Data Set' https://www.mdpi.com/2079-9292/11/8/1199/htm that were diagnosed with condition(s) of: Non-small cell carcinoma of lung,TNM stage 1,  Non-small cell lung cancer, Suspected lung cancer.  Data hosted by: aced-manchester~aced-manchester
Prostate_Cancer~35488~Patients from 'Coherent Data Set' https://www.mdpi.com/2079-9292/11/8/1199/htm that were diagnosed with condition(s) of: Metastasis from malignant tumor of prostate, Neoplasm of prostate, arcinoma in situ of prostate.  Data hosted by: aced-stanford~aced-stanford""".split('\n')  # noqa E501

    for line in discovery_descriptions:
        if len(line) == 0:
            continue
        (name, _subjects_count, description, location,) = line.split('~')
        gen3_discovery = {'tags': [
            {"name": program, "category": "Program"},
            {"name": f"aced_{name}", "category": "Study Registration"},
            {"name": location, "category": "Study Location"},

        ], 'name': name, 'full_name': name, 'study_description': description}

        guid = f"aced_{name}"

        gen3_discovery['commons'] = "ACED"
        gen3_discovery['commons_name'] = "ACED Commons"
        # TODO - read this value for commons_url from some other, more dynamic source
        gen3_discovery['commons_url'] = 'staging.aced-idp.org'
        gen3_discovery['__manifest'] = 0
        gen3_discovery['_research_subject_count'] = int(_subjects_count)
        gen3_discovery['_unique_id'] = guid
        gen3_discovery['study_id'] = guid
        discoverable_data = dict(_guid_type="discovery_metadata", gen3_discovery=gen3_discovery)
        discovery_client.create(guid, discoverable_data, aliases=None, overwrite=True)
        print(f"Added {name}")
